from typing import Tuple

import numba as nb
import numpy as np
from scipy.special import gamma, ndtr

from beetroots.modelling.priors import smooth_indicator_prior

GAMMA_1_4 = gamma(1 / 4)


@nb.njit()
def sample_generalized_gaussian(
    alpha: float,
    a_alpha: float,
    size: Tuple[int, int],
    seed: int,
) -> np.ndarray:
    r"""sample from a generalized Gaussian distribution of pdf:

    .. math::

        p(\theta) = \frac{2}{\delta \Gamma(1/4)} \exp \left\{- \left( \frac{\theta}{\delta}\right)^4\right\}

    with here :math:`\delta = 1/A(\alpha)`.
    See :cite:t:`nardonSimulationTechniquesGeneralized2009` for more details.

    Parameters
    ----------
    alpha : float
        power used in the distribution. alpha=2 is equivalent to the Gaussian distribution.
    a_alpha : float
        scaling parameter
    size : Tuple[int, int]
        number of samples to draw
    seed : int
        random seed for sampling reproducibility

    Returns
    -------
    np.ndarray
        array of samples
    """
    np.random.seed(seed)
    z = np.random.gamma(shape=1 / alpha, scale=(1 / a_alpha) ** alpha, size=size)
    x = np.random.binomial(n=1, p=0.5, size=size) * 2 - 1
    return x * z ** (1 / alpha)


@nb.njit()
def sample_smooth_indicator(
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    indicator_margin_scale: float,
    size: Tuple[int, int],
    seed: int,
) -> np.ndarray:
    r"""algorithm to draw iid samples from the smooth indicator prior, detailed in Appendix B in :cite:t:`paludEfficientSamplingNon2023`

    Parameters
    ----------
    lower_bounds : np.ndarray
        vector of lower bounds for each of the D physical parameters
    upper_bounds : np.ndarray
        vector of upper bounds for each of the D physical parameters
    indicator_margin_scale : float
        scaling parameter, quantifies how much values out of the validity intervals are penalized
    size : Tuple[int, int]
        number of samples to draw
    seed : int
        random seed, for sampling reproducibility

    Returns
    -------
    np.ndarray
        iid samples drawn from the smooth uniform distribution
    """
    np.random.seed(seed)
    k_mtm, D = size

    Z_unif = 2 * (upper_bounds - lower_bounds) / (indicator_margin_scale * GAMMA_1_4)
    Z_tot = 1 + Z_unif
    w_unif = Z_unif / Z_tot  # (D,)

    z1 = np.zeros((k_mtm, D))
    x_unif = np.zeros((k_mtm, D))
    for d in range(D):
        z1[:, d] = np.random.binomial(1, w_unif[d], size=k_mtm)
        x_unif[:, d] = np.random.uniform(lower_bounds[d], upper_bounds[d], size=k_mtm)

    x_gg = sample_generalized_gaussian(4.0, 1 / indicator_margin_scale, size, seed)
    x_gg = np.where(
        x_gg < 0,
        x_gg + np.expand_dims(lower_bounds, 0),
        x_gg + np.expand_dims(upper_bounds, 0),
    )
    return np.where(z1 == 1, x_unif, x_gg)


@nb.njit()
def get_neighboring_pixels(
    current_Theta: np.ndarray, list_edges: np.ndarray, idx_pix: int
) -> np.ndarray:
    r"""returns the array of current values of the neighboring pixels of a specified pixel

    Parameters
    ----------
    current_Theta : np.ndarray
        current iterate
    list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    idx_pix : int
        index of a pixel

    Returns
    -------
    np.ndarray
        array of current values of the neighboring pixels
    """
    list_edges_n = list_edges[
        (list_edges[:, 0] == idx_pix) | (list_edges[:, 1] == idx_pix)
    ]
    list_neighbors_idx = list_edges_n.flatten()
    list_neighbors_idx = list_neighbors_idx[list_neighbors_idx != idx_pix]
    return current_Theta[list_neighbors_idx, :]  # (n_neighbors, D)


@nb.njit()
def sample_conditional_spatial_and_indicator_prior(
    current_Theta: np.ndarray,
    spatial_list_edges: np.ndarray,
    spatial_weights: np.ndarray,
    indicator_lower_bounds: np.ndarray,
    indicator_upper_bounds: np.ndarray,
    indicator_indicator_margin_scale: float,
    idx_pix: np.ndarray,
    k_mtm: int,
    seed: int,
) -> np.ndarray:
    r"""draw ``k_mtm`` iid samples for each candidate in the `ìdx_pix`` array.

    Parameters
    ----------
    current_Theta : np.ndarray
        current iterate
    spatial_list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    spatial_weights : np.ndarray
        regularization weights of the spatial prior
    indicator_lower_bounds : np.ndarray
        vector of lower bounds for the D physical parameters
    indicator_upper_bounds : np.ndarray
        vector of upper bounds for the D physical parameters
    indicator_indicator_margin_scale : float
        scaling parameter, quantifies how much values out of the validity intervals are penalized
    idx_pix : np.ndarray
        array of indices of pixels for which candidates are to be proposed
    k_mtm : int
        number of candidates in the MTM kernel
    seed : int
        random seed, used for sampling reproducibility

    Returns
    -------
    np.ndarray
        samples drawn from the conditional prior
    """
    np.random.seed(seed)
    (N, D) = current_Theta.shape
    n_pix = idx_pix.size * 1

    samples = np.zeros((n_pix, k_mtm, D))
    i = 0
    for idx_1_pix in idx_pix:
        # * sample from around neighbors
        neighbors = get_neighboring_pixels(current_Theta, spatial_list_edges, idx_1_pix)
        N_neighbors = neighbors.shape[0]

        if N_neighbors > 0:
            Theta = np.zeros((k_mtm, D))

            for k in range(k_mtm):
                # select which combination f neighbors is going to be used
                arr_use_neighbors = np.random.binomial(n=1, p=0.5, size=N_neighbors)
                while np.max(arr_use_neighbors) == 0:
                    arr_use_neighbors = np.random.binomial(n=1, p=0.5, size=N_neighbors)

                used_neighbors = neighbors[arr_use_neighbors == 1]
                N_used_neighbors = used_neighbors.shape[0]

                # sigma_mtm_eff = 1 / (2 * np.sqrt(N_neighbors * spatial_weights))  # (D,)
                sigma_mtm_eff = 1 / (
                    2 * np.sqrt(N_used_neighbors * spatial_weights)
                )  # (D,)

                # initialize array of candidates

                mean = np.zeros((1, D))
                for d in range(D):
                    mean[0, d] = np.mean(used_neighbors[:, d])
                    # mean[d] = np.mean(neighbors[:, d])

                repeat = True
                n_repeats = 0
                n_repeats_tot = 0
                while repeat:
                    # * step 1 : generate candidates from spatial prior only
                    Theta_cand = mean + sigma_mtm_eff * np.random.standard_normal(
                        size=(1, D)
                    )

                    # * step 2 : accept or reject with combination of spatial
                    # * and indicator priors
                    p_Theta = np.exp(
                        -smooth_indicator_prior.penalty_one_pix(
                            Theta_cand,
                            indicator_lower_bounds,
                            indicator_upper_bounds,
                            indicator_indicator_margin_scale,
                        )
                    )  # (1,)
                    u = np.random.uniform(0, 1)
                    if u <= p_Theta[0]:
                        Theta[k] += Theta_cand[0]
                        repeat = False

                    n_repeats += 1
                    n_repeats_tot += 1

                    if n_repeats >= 10:
                        # select which combination f neighbors is going to be used
                        arr_use_neighbors = np.random.binomial(
                            n=1, p=0.5, size=N_neighbors
                        )
                        while np.max(arr_use_neighbors) == 0:
                            arr_use_neighbors = np.random.binomial(
                                n=1, p=0.5, size=N_neighbors
                            )

                        used_neighbors = neighbors[arr_use_neighbors == 1]
                        N_used_neighbors = used_neighbors.shape[0]

                        # sigma_mtm_eff = 1 / (2 * np.sqrt(N_neighbors * spatial_weights))  # (D,)
                        sigma_mtm_eff = 1 / (
                            2 * np.sqrt(N_used_neighbors * spatial_weights)
                        )  # (D,)

                        # initialize array of candidates

                        mean = np.zeros((1, D))
                        for d in range(D):
                            mean[0, d] = np.mean(used_neighbors[:, d])
                            # mean[d] = np.mean(neighbors[:, d])

                        n_repeats = 0

                    assert n_repeats_tot < 1_000

            samples[i, :, :] = Theta * 1  # (k_mtm, D)

        else:
            samples[i, :, :] = sample_smooth_indicator(
                indicator_lower_bounds,
                indicator_upper_bounds,
                indicator_indicator_margin_scale,
                size=(k_mtm, D),
                seed=seed,
            )
        i += 1

    return samples


@nb.njit()
def compute_sum_subsets_norms(dists: np.ndarray) -> np.ndarray:
    r"""computes the sums of the elements of each subset of neighbors (including the empty set, excluding the full set of neighbors) by exploiting the tree structure of the subsets.

    used in step 3 of ``compute_nlratio_prior_proposal``

    Parameters
    ----------
    dists : np.ndarray of shape (N_neighbors, k_mtm, D)
        _description_

    Returns
    -------
    np.ndarray
        _description_
    """
    N_neighbors, k_mtm, D = dists.shape
    sums_ = np.zeros((2**N_neighbors - 1, k_mtm, D))

    if N_neighbors == 1:
        # sums_ : (1, k_mtm + 1, D)
        return sums_

    if N_neighbors == 2:
        # sums_ : (3, k_mtm + 1, D)
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = dists[1, :] * 1
        return sums_

    if N_neighbors == 3:
        # sums_ : (7, k_mtm + 1, D)
        # 1st neighbor
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = sums_[1, :] + dists[1, :]
        sums_[3, :] = sums_[1, :] + dists[2, :]
        # 2nd neighbor
        sums_[4, :] = dists[1, :] * 1
        sums_[5, :] = sums_[4, :] + dists[2, :]
        # 3rd neighbor
        sums_[6, :] = dists[2, :] * 1
        return sums_

    if N_neighbors == 4:
        # sums_ : (15, k_mtm + 1, D)
        # 1st neighbor
        sums_[1, :] = dists[0, :] * 1
        sums_[2, :] = sums_[1, :] + dists[1, :]
        sums_[3, :] = sums_[2, :] + dists[2, :]
        sums_[4, :] = sums_[2, :] + dists[3, :]
        sums_[5, :] = sums_[1, :] + dists[2, :]
        sums_[6, :] = sums_[5, :] + dists[3, :]
        sums_[7, :] = sums_[1, :] + dists[3, :]
        # 2nd neighbor
        sums_[8, :] = dists[1, :] * 1
        sums_[9, :] = sums_[8, :] + dists[2, :]
        sums_[10, :] = sums_[9, :] + dists[3, :]
        sums_[11, :] = sums_[8, :] + dists[3, :]
        # 3rd neighbor
        sums_[12, :] = dists[2, :] * 1
        sums_[13, :] = sums_[12, :] + dists[3, :]
        # 4th neighbor
        sums_[14, :] = dists[3, :] * 1
        return sums_

    else:
        assert np.min(sums_) > 0
        return sums_


@nb.njit(fastmath=True)
def numba_logsumexp_stable(x: np.ndarray) -> float:
    r"""stable numba implementation of the ``logsumexp`` function. Used in step 4 of ``compute_nlratio_prior_proposal``.

    Parameters
    ----------
    x : np.ndarray
        array of real values

    Returns
    -------
    float
        result of the ``logsumexp`` function
    """
    x_max = np.max(x)
    res = 0.0
    for j in range(x.size):
        res += np.exp(x[j] - x_max)
    res = np.log(res) + x_max
    return res


@nb.njit()
def compute_nlratio_prior_proposal(
    new_Theta: np.ndarray,
    spatial_list_edges: np.ndarray,
    spatial_weights: np.ndarray,
    idx_pix: np.ndarray,
    candidates_pix: np.ndarray,
) -> np.ndarray:
    """evaluates the negative log ratio of prior proposal in the MTM-chromatic Gibbs kernel in ``MySampler``

    Note
    ----
    The function works in four steps :

    * step 1: get the values of the neighbors of each pixel in ``idx_pix`` using ``get_neighboring_pixels``
    * step 2: compute the vector distances between the candidate and the pixel neighbors
    * step 3: compute list of sums of subsets using ``compute_sum_subsets_norms``
    * step 4: evaluate ``LogSumExp`` to obtain the log pdf of the Gaussian mixture, using the stable ``numba`` implementation.

    Parameters
    ----------
    new_Theta : np.ndarray
        iterate
    spatial_list_edges : np.ndarray
        list of edges, defined as pairs of pixel indices. Generated by the class:`SpatialPrior`.
    spatial_weights : np.ndarray
        regularization weights of the spatial prior
    idx_pix : np.ndarray of shape (n_pix,)
        array of indices of pixels for which candidates are to be proposed
    candidates_pix : np.ndarray of shape (n_pix, k_mtm, D)
        array of the candidate values

    Returns
    -------
    np.ndarray of shape (n_pix, k_mtm)
        negative log ratio of prior proposal
    """
    n_pix, k_mtm, D = candidates_pix.shape

    nl_ratio = np.zeros((n_pix, k_mtm))
    for i in range(n_pix):
        idx_1_pix = idx_pix[i]

        # * step 1: get neighbors
        neighbors = get_neighboring_pixels(
            new_Theta, spatial_list_edges, idx_1_pix
        )  # (N_neighbors, D)
        N_neighbors = neighbors.shape[0]

        # * step 2: compute distances between candidate and its neighbors
        dists = (
            np.expand_dims(candidates_pix[i], 0) - np.expand_dims(neighbors, 1)
        ) ** 2  # (N_neighbors, k_mtm, D)
        assert dists.shape == (N_neighbors, k_mtm, D)

        # * step 3: compute list of sums of subsets
        sums_norms = compute_sum_subsets_norms(
            dists
        )  # (2 ** N_neighbors - 1, k_mtm, D)
        assert sums_norms.shape == (2**N_neighbors - 1, k_mtm, D)

        # * step 4: LogSumExp
        inter_ = 2 * np.expand_dims(np.expand_dims(spatial_weights, 0), 0) * sums_norms
        # (2 ** N_neighbors - 1, k_mtm, D)
        for k in range(k_mtm):
            for d in range(D):
                nl_ratio[i, k] += numba_logsumexp_stable(inter_[:, k, d])

    return nl_ratio



import abc
import os
from typing import List, Union

import numpy as np
import pandas as pd

from beetroots.modelling.priors.abstract_prior import PriorProbaDistribution
from beetroots.modelling.priors.spatial_prior_params import SpatialPriorParams


def build_list_edges(
    df: pd.DataFrame,
    use_next_nearest_neighbors: bool = True,
) -> np.ndarray:
    r"""builds the list of edges necessary in the rest of the functions of this file

    Parameters
    ----------
    df : pandas.DataFrame
        observation map data. The index needs to contain 2 columns (one for the x-coordinate of the pixel and one for its y-coordinate)
    use_next_nearest_neighbors : bool, optional
        wether or not to use the next nearest neighbors, i.e., in diagonal, by default True

    Returns
    -------
    list_edges : numpy.array of shape (-1, 2)
        list of 2 by 2 neighboring relations without duplicates (each pair is ordered by lowest id to highest id)
    """
    df_clusters = df.copy()
    df_clusters = df_clusters.reset_index().set_index("idx")
    df_clusters["clusters"] = 0
    df_clusters = df_clusters["clusters"]

    list_edges = []
    list_considered_neighbours = (
        [(1, 0), (0, 1), (1, 1), (-1, 1)]
        if use_next_nearest_neighbors
        else [(1, 0), (0, 1)]
    )
    for (x, y) in list(df.index):
        idx = df.at[(x, y), "idx"]
        cluster_current = df_clusters.at[idx]
        for (delta_x, delta_y) in list_considered_neighbours:
            x2 = x + delta_x
            y2 = y + delta_y
            if (x2, y2) in df.index:
                idx2 = df.at[(x2, y2), "idx"]
                if df_clusters.at[idx2] == cluster_current:
                    list_edges.append([idx, idx2])
    list_edges = np.array(list_edges)
    # list_edges = list_edges[list_edges[:, 0].argsort()]
    print(f"total number of edges in img graph : {list_edges.size // 2}")
    return list_edges


class SpatialPrior(PriorProbaDistribution):
    r"""Abstract Base Class for a spatial regularizing prior"""

    __slots__ = (
        "D",
        "N",
        "use_next_nearest_neighbors",
        "list_edges",
        "dict_sites",
        "initial_weights",
        "weights",
    )

    def __init__(
        self,
        spatial_prior_params: SpatialPriorParams,
        cloud_name: str,
        D: int,
        N: int,
        df: pd.DataFrame,
        list_idx_sampling: List[int],
    ) -> None:
        super().__init__(D, N)

        # list of neighboring relations
        self.use_next_nearest_neighbors = (
            spatial_prior_params.use_next_nearest_neighbors
        )
        r"""bool: wether to use the next nearest neighbors (i.e., neighbors in diagonal)"""

        self.list_edges = build_list_edges(df, self.use_next_nearest_neighbors)
        r"""np.ndarray: set of edges in the graph induced by the spatial regularization"""

        self.dict_sites = self.build_sites(df)
        """dict[int, np.ndarray]: sites of the graph induced the spatial regularization. A site is a set of nodes that are conditionally independent."""

        # one weight per dimension
        initial_weights = spatial_prior_params.initial_regu_weights * 1

        if initial_weights is not None:
            if isinstance(initial_weights, list):
                initial_weights = np.array(initial_weights)

            initial_weights = initial_weights[list_idx_sampling] * 1
            weights = initial_weights[list_idx_sampling] * 1

            # assert self.initial_weights.shape == (
            #     self.D,
            # ), f"{self.initial_weights.shape} should be size {self.D}"
        else:
            initial_weights = np.ones((D,))
            weights = np.ones((D,))

        self.initial_weights = initial_weights
        r"""np.ndarray of shape (D,): initial weights of the spatial regularization :math:`\tau`"""

        self.weights = weights
        r"""np.ndarray of shape (D,): current value of weights of the spatial regularization :math:`\tau`"""

        return

    def build_sites(self, df: pd.DataFrame) -> dict[int, np.ndarray]:
        """builds the site from the DataFrame of positions

        Parameters
        ----------
        df : pd.DataFrame
            positions of the pixels

        Returns
        -------
        dict[int, np.ndarray]
            set of sites and corresponding pixels. Forms a partition of the full set of pixels.
        """
        if self.use_next_nearest_neighbors:
            dict_sites_raw = {i: [] for i in range(4)}
            for (x, y) in list(df.index):
                idx_site = 2 * (x % 2) + y % 2
                dict_sites_raw[idx_site].append(df.at[(x, y), "idx"])
        else:
            dict_sites_raw = {i: [] for i in range(2)}
            for (x, y) in list(df.index):
                idx_site = (x + y) % 2
                dict_sites_raw[idx_site].append(df.at[(x, y), "idx"])

        dict_sites = {}
        for k, list_idx in dict_sites_raw.items():
            dict_sites[k] = np.sort(list_idx)

        return dict_sites

    @abc.abstractmethod
    def neglog_pdf(
        self,
        Theta: np.ndarray,
        with_weights: bool = True,
    ) -> np.ndarray:
        r"""computes the negative log pdf (defined up to some multiplicative constant)

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            set of D maps on which we want to evaluate the neg log prior

        Returns
        -------
        neglog_p : np.array of shape (D,)
            set of the D neg log priors evaluation
        """
        pass

    @abc.abstractmethod
    def neglog_pdf_one_pix(
        self, Theta: np.ndarray, n: int, with_weights: bool = True
    ) -> np.ndarray:
        r"""computes the negative log of the pdf (defined up to some multiplicative constant) in the neighborhood of one pixel

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            set of D maps on which we want to evaluate the neg log prior

        Returns
        -------
        neglog_p : np.array of shape (D,)
            set of the D neg log priors evaluation
        """
        pass

    @abc.abstractmethod
    def gradient_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        r"""Computes the gradient of the spatial regularization

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            current iterate

        Returns
        -------
        np.array of shape (N, D)
            gradient of the spatial regularization
        """
        pass

    @abc.abstractmethod
    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        r"""Computes the diagonal of the hessian of the spatial regularization

        Parameters
        ----------
        Theta : np.array of shape (N, D)
            current iterate

        Returns
        -------
        np.array of shape (N, D)
            diagonal of the hessian of the spatial regularization
        """
        pass



import abc


class ApproachType(abc.ABC):
    r"""abstract class to define the optimization method for the likelihood parameter adjustment"""

    @abc.abstractmethod
    def optimization(self):
        pass

    @abc.abstractmethod
    def plot_cost_map(self):
        pass



import numpy as np

from beetroots.approx_optim.forward_map.abstract_forward_map import (
    ApproxOptimForwardMap,
)
from beetroots.simulations.astro.forward_map.abstract_nn import SimulationNN


class ApproxOptimNN(ApproxOptimForwardMap):
    r"""handles the generation of a dataset of :math:`\log_{10} f_{\ell}(\theta)` values for a neural network forward map"""

    def compute_log10_f_Theta(
        self,
        dict_forward_model: dict,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
    ) -> np.ndarray:
        simulation = SimulationNN()
        simulation.list_lines_fit = self.list_lines * 1

        scaler, forward_map = simulation.setup_forward_map(
            forward_model_name=dict_forward_model["forward_model_name"],
            force_use_cpu=dict_forward_model["force_use_cpu"],
            dict_fixed_params=dict_forward_model["fixed_params"],
            dict_is_log_scale_params=dict_forward_model["is_log_scale_params"],
        )

        lower_bounds = scaler.from_lin_to_scaled(
            lower_bounds_lin.reshape((1, self.D)),
        ).flatten()  # (D,)
        upper_bounds = scaler.from_lin_to_scaled(
            upper_bounds_lin.reshape((1, self.D)),
        ).flatten()  # (D,)

        # restrict bounds to the unfixed parameters
        lower_bounds = lower_bounds[self.list_idx_sampling] * 1
        upper_bounds = upper_bounds[self.list_idx_sampling] * 1

        dict_fixed_params_scaled = simulation.scale_dict_fixed_params(
            scaler, dict_forward_model["fixed_params"]
        )
        print(dict_fixed_params_scaled)

        Theta = np.zeros((self.K**self.D_sampling, self.D))
        Theta[:, self.list_idx_sampling] = self.sample_theta(lower_bounds, upper_bounds)

        for i, v in enumerate(dict_fixed_params_scaled.values()):
            if v is not None:
                assert np.allclose(Theta[:, i], 0.0), Theta[:10]
                Theta[:, i] = v * 1

        print(f"Theta.shape = {Theta.shape}")

        # the division is to get log in base 10
        print("starting computation of set of log10 f(theta) values")

        log10_f_Theta = forward_map.evaluate_log(Theta) / np.log(10)
        assert log10_f_Theta.shape == (
            self.N_samples_theta,
            self.L,
        ), f"log10_f_Theta.shape = {log10_f_Theta.shape}, (self.N_samples_theta, self.L) = {(self.N_samples_theta, self.L)}"

        print(f"log10_f_Theta.shape = {log10_f_Theta.shape}")
        print("computation of set of log10 f(theta) values done")

        return log10_f_Theta



import multiprocessing as mp
import os
import time
import warnings
from concurrent.futures import ProcessPoolExecutor
from typing import Dict, List, Tuple, Union

import h5py
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.inversion.results.utils.mc_utils import ess, histograms
from beetroots.space_transform.abstract_transform import Scaler


class ResultsMC(ResultsUtil):

    __slots__ = (
        "model_name",
        "chain_type",
        "path_img",
        "path_data_csv_out_mcmc",
        "max_workers",
        "N_MCMC",
        "T_MC",
        "T_BI",
        "freq_save",
        "effective_T_BI",
        "lower_bounds_lin",
        "upper_bounds_lin",
        "N",
        "D",
        "list_names",
    )

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        max_workers: int,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        N: int,
        list_idx_sampling: List,
        list_fixed_values_scaled: List,
        lower_bounds_lin: Union[np.ndarray, List[float]],
        upper_bounds_lin: Union[np.ndarray, List[float]],
        list_names: List[str],
    ):
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.chain_type = chain_type
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc
        self.max_workers = max_workers

        self.N_MCMC = N_MCMC
        self.T_MC = T_MC
        self.T_BI = T_BI
        self.freq_save = freq_save
        self.effective_T_BI = T_BI // freq_save

        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)
        if isinstance(upper_bounds_lin, list):
            upper_bounds_lin = np.array(upper_bounds_lin)

        self.lower_bounds_lin = lower_bounds_lin
        self.upper_bounds_lin = upper_bounds_lin

        self.list_idx_sampling = list_idx_sampling
        self.list_fixed_values_scaled = list_fixed_values_scaled

        self.N = N
        self.D = upper_bounds_lin.size
        self.D_sampling = len(self.list_idx_sampling)

        self.list_names = list_names

    def read_data(self):
        pass

    def create_folders(self) -> Tuple[str]:
        folder_path_mc = f"{self.path_img}/mc"

        folder_path_1D = f"{folder_path_mc}/{self.model_name}_1D"
        folder_path_1D_chain = f"{folder_path_1D}/chains"
        folder_path_1D_hist = f"{folder_path_1D}/hist"

        folder_path_2D = f"{folder_path_mc}/{self.model_name}_2D"
        folder_path_2D_chain = f"{folder_path_2D}/chains"
        folder_path_2D_hist = f"{folder_path_2D}/hist"
        folder_path_2D_proba = f"{folder_path_2D}/proba_contours"

        for path_ in [
            folder_path_mc,
            folder_path_1D,
            folder_path_1D_chain,
            folder_path_1D_hist,
            folder_path_2D,
            folder_path_2D_chain,
            folder_path_2D_hist,
            folder_path_2D_proba,
        ]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return (
            folder_path_1D_chain,
            folder_path_1D_hist,
            folder_path_2D_chain,
            folder_path_2D_hist,
            folder_path_2D_proba,
        )

    def full_mc_analysis(
        self,
        scaler: Scaler,
        Theta_true_scaled_full: np.ndarray,
        list_mcmc_folders: List[str],
        plot_ESS: bool,
        plot_1D_chains: bool,
        plot_2D_chains: bool,
        plot_comparisons_yspace: bool,
        #
        folder_path_1D_chain: str,
        folder_path_1D_hist: str,
        folder_path_2D_chain: str,
        folder_path_2D_hist: str,
        folder_path_2D_proba: str,
        #
        point_challenger: Dict = {},
        list_CI: List[int] = [],
    ) -> None:
        global _one_pixel_mmse_ic_extraction

        list_fixed_values_lin = np.zeros((1, self.D))
        for d, value in enumerate(self.list_fixed_values_scaled):
            if value is not None:
                list_fixed_values_lin[0, d] = value * 1

        list_fixed_values_lin = scaler.from_scaled_to_lin(
            list_fixed_values_lin,
        ).flatten()  # (D,)

        # list_fixed_values_lin = [
        #     v if d not in self.list_idx_sampling else None
        #     for d, v in enumerate(list_fixed_values_lin)
        # ]

        len_mc = (self.T_MC - self.T_BI) // self.freq_save

        def _one_pixel_mmse_ic_extraction(dict_input: dict):
            """for one pixel n, performs:
            - MMSE and credibility interval extraction
            - ESS computation
            - plot 1D histograms
            - plot 2D histograms
            """
            n = dict_input["n"]
            Theta_n_true = dict_input["Theta_n_true"]

            # read data
            list_Theta_n_lin_full = np.zeros((self.N_MCMC, len_mc, self.D))
            for d in range(self.D):
                if d not in self.list_idx_sampling:
                    list_Theta_n_lin_full[:, :, d] = list_fixed_values_lin[d] * 1

            for seed, mc_path in enumerate(list_mcmc_folders):
                with h5py.File(mc_path, "r") as f:
                    list_Theta_n_lin_sub = np.array(
                        f["list_Theta"][self.effective_T_BI :, n, :]
                    )

                for idx, d in enumerate(self.list_idx_sampling):
                    list_Theta_n_lin_full[seed, :, d] = list_Theta_n_lin_sub[:, idx]

            # * MMSE and IC estimators
            list_Theta_n_lin_full_flatter = list_Theta_n_lin_full.reshape(
                (self.N_MCMC * (len_mc), self.D)
            )

            # compute percentiles
            dict_ci_per = {
                ci: {"lower": (100 - ci) / 2, "upper": 100 - (100 - ci) / 2}
                for ci in list_CI
            }
            list_per = list([dict_ci_per[ci]["lower"] for ci in list_CI])
            list_per += list([dict_ci_per[ci]["upper"] for ci in list_CI])
            list_per.sort()
            dict_per = {
                per: np.percentile(list_Theta_n_lin_full_flatter, per, axis=0)  # (D,)
                for per in list_per
            }

            # compute MMSE
            list_Theta_n_scaled_full_flatter = scaler.from_lin_to_scaled(
                list_Theta_n_lin_full_flatter,
            )

            # for d in range(self.D):
            #     if self.list_fixed_values_scaled[d] is not None:
            #         list_Theta_n_scaled_flatter[:, d] += self.list_fixed_values_scaled[d]

            Theta_n_MMSE_scaled = np.mean(
                list_Theta_n_scaled_full_flatter, axis=0
            )  # (D,)
            Theta_n_MMSE_lin = scaler.from_scaled_to_lin(
                Theta_n_MMSE_scaled.reshape((1, self.D))
            ).flatten()  # (D,)

            assert Theta_n_MMSE_lin.shape == (
                self.D,
            ), f"shape {Theta_n_MMSE_lin.shape}, should be {(self.D,)}"

            # create and save dataset of MMSE and IC
            df_estim = pd.DataFrame()
            df_estim["n"] = n * np.ones((self.D,), dtype=np.int32)
            df_estim["d"] = np.arange(self.D)
            df_estim["Theta_MMSE"] = Theta_n_MMSE_lin * 1
            for per in list_per:
                df_estim[f"per_{per:.1f}".replace(".", "p")] = dict_per[per]

            # in order to avoid re-writing multiple times the header because of
            # parallel writing, force a delay to favor n = 0 to be written
            # first with header
            path_overall_results = f"{self.path_data_csv_out_mcmc}/"
            path_overall_results += f"estimation_Theta_{self.model_name}.csv"

            if n == 0:
                df_estim.to_csv(path_overall_results, mode="w")
            else:
                while not (os.path.exists(path_overall_results)):
                    time.sleep(0.5)

                df_estim.to_csv(
                    path_overall_results,
                    mode="a",
                    header=not (os.path.exists(path_overall_results)),
                )

            del df_estim

            # *  index of first element st true val btw [MC first val, elt] or
            # * [elt, MC first val]
            if Theta_n_true is not None:
                first_elt_arr = -np.ones((self.N_MCMC, self.D_sampling))
                for seed in range(self.N_MCMC):
                    for idx_d, d in enumerate(self.list_idx_sampling):
                        if list_Theta_n_lin_full[seed, 0, d] < Theta_n_true[d]:
                            (idx,) = np.where(
                                list_Theta_n_lin_full[seed, :, d] >= Theta_n_true[d],
                            )
                        else:
                            (idx,) = np.where(
                                list_Theta_n_lin_full[seed, :, d] <= Theta_n_true[d],
                            )

                        if idx.size > 0:
                            first_elt_arr[seed, idx_d] = idx[0]

                list_dict = [
                    {
                        "seed": seed,
                        "n": n,
                        "d": d,
                        "first_elt_valid_mc": int(first_elt_arr[seed, idx_d]),
                    }
                    for seed in range(self.N_MCMC)
                    for idx_d, d in enumerate(self.list_idx_sampling)
                ]
                df_first_elt_valid_mc = pd.DataFrame.from_records(list_dict)

                path_file = f"{self.path_data_csv_out_mcmc}/"
                path_file += f"first_elt_valid_mc_{self.model_name}.csv"
                if n == 0:
                    df_first_elt_valid_mc.to_csv(
                        path_file,
                        mode="w",
                    )
                else:
                    while not (os.path.exists(path_file)):
                        time.sleep(0.5)

                    df_first_elt_valid_mc.to_csv(
                        path_file,
                        mode="a",
                        header=not (os.path.exists(path_file)),
                    )

                del df_first_elt_valid_mc

            # * ESS
            if plot_ESS:
                list_Theta_n_scaled_full = list_Theta_n_scaled_full_flatter.reshape(
                    (self.N_MCMC, len_mc, self.D)
                )
                list_dict_output = []
                for d in self.list_idx_sampling:
                    ess_ = ess.compute_ess(list_Theta_n_scaled_full[:, :, d])
                    list_dict_output.append(
                        {
                            "n": n,
                            "d": d,
                            "seed": "overall",
                            "model_name": self.model_name,
                            "ess": ess_,
                        }
                    )

                df_ess_nd = pd.DataFrame.from_records(list_dict_output)

                path_file = f"{self.path_data_csv_out_mcmc}/"
                path_file += f"estimation_ESS_{self.model_name}.csv"
                if n == 0:
                    df_ess_nd.to_csv(path_file, mode="w")

                else:
                    while not (os.path.exists(path_file)):
                        time.sleep(0.5)

                    df_ess_nd.to_csv(
                        path_file,
                        mode="a",
                        header=not (os.path.exists(path_file)),
                    )

                del df_ess_nd

            # * 1D histograms
            if plot_1D_chains:
                for d in self.list_idx_sampling:
                    true_val = Theta_n_true[d] if Theta_n_true is not None else None

                    histograms.plot_1D_chain(
                        list_Theta_n_lin_full_flatter[:, d],
                        n,
                        d,
                        folder_path_1D_chain,
                        f"MC component {self.list_names[d]} of pixel {n}",
                        self.lower_bounds_lin,
                        self.upper_bounds_lin,
                        self.N_MCMC,
                        self.T_MC,
                        self.T_BI,
                        true_val,
                    )

                    histograms.plot_1D_hist(
                        list_Theta_n_lin_full_flatter[:, d],
                        n,
                        d,
                        folder_path_1D_hist,
                        title=f"hist. of {self.list_names[d]} of pixel {n}",
                        lower_bounds_lin=self.lower_bounds_lin,
                        upper_bounds_lin=self.upper_bounds_lin,
                        seed=None,
                        estimator=Theta_n_MMSE_lin[d],
                        true_val=true_val,
                    )

            # * 2D histograms
            if plot_2D_chains and self.D > 1:
                for idx_d1, d1 in enumerate(self.list_idx_sampling):
                    for d2 in self.list_idx_sampling[idx_d1 + 1 :]:
                        if Theta_n_true is not None:
                            true_val = Theta_n_true[[d1, d2]] * 1
                        else:
                            true_val = None

                        histograms.plot_2D_hist(
                            list_Theta_n_lin_full_flatter[:, [d1, d2]],
                            n,
                            d1,
                            d2,
                            self.model_name,
                            folder_path_2D_hist,
                            self.list_names,
                            self.lower_bounds_lin,
                            self.upper_bounds_lin,
                            Theta_MMSE=Theta_n_MMSE_lin[[d1, d2]],
                            true_val=true_val,
                            point_challenger=point_challenger,
                        )

                        try:
                            histograms.plot_2D_proba_contours(
                                list_Theta_n_lin_full_flatter[:, [d1, d2]],
                                n,
                                d1,
                                d2,
                                self.model_name,
                                folder_path_2D_proba,
                                self.list_names,
                                self.lower_bounds_lin,
                                self.upper_bounds_lin,
                                Theta_MMSE=Theta_n_MMSE_lin[[d1, d2]],
                                true_val=true_val,
                                point_challenger=point_challenger,
                            )
                        except:
                            msg = "Issue with proba contour plot for (n, d1, d2) = "
                            msg += f"({n}, {d1}, {d2})"
                            print(msg)

            return

        # * global part of the function
        if Theta_true_scaled_full is not None:
            Theta_true_lin = scaler.from_scaled_to_lin(Theta_true_scaled_full)
        else:
            Theta_true_lin = None

        list_params = [
            {
                "n": n,
                "Theta_n_true": Theta_true_lin[n]
                if Theta_true_lin is not None
                else None,
            }
            for n in range(self.N)
        ]

        # ? The parallel execution may fail on mac, even with the mp_context
        # ? argument. As I can't correct the error, in case of fail, I perform
        # ? the extraction in series, which is much slower.
        try:
            with ProcessPoolExecutor(
                max_workers=self.max_workers, mp_context=mp.get_context("fork")
            ) as p:
                _ = list(
                    tqdm(
                        p.map(_one_pixel_mmse_ic_extraction, list_params),
                        total=self.N,
                    )
                )
        except:
            warnings.warn(
                "The parallel pixel-wise result extraction failed. Extracting in series instead."
            )
            for params in tqdm(list_params):
                _one_pixel_mmse_ic_extraction(params)

        return

    def main(
        self,
        scaler: Scaler,
        Theta_true_scaled_full: np.ndarray,
        list_mcmc_folders: List[str],
        plot_ESS: bool,
        plot_1D_chains: bool,
        plot_2D_chains: bool,
        plot_comparisons_yspace: bool,
        point_challenger: Dict = {},
        list_CI: List[int] = [],
    ):
        (
            folder_path_1D_chain,
            folder_path_1D_hist,
            folder_path_2D_chain,
            folder_path_2D_hist,
            folder_path_2D_proba,
        ) = self.create_folders()

        self.full_mc_analysis(
            scaler,
            Theta_true_scaled_full,
            list_mcmc_folders,
            plot_ESS,
            plot_1D_chains,
            plot_2D_chains,
            plot_comparisons_yspace,
            #
            folder_path_1D_chain,
            folder_path_1D_hist,
            folder_path_2D_chain,
            folder_path_2D_hist,
            folder_path_2D_proba,
            #
            point_challenger=point_challenger,
            list_CI=list_CI,
        )
        return



from typing import Optional

import numba
import numpy as np

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


@numba.njit()
def compute_laplacian(Theta: np.ndarray, list_edges: np.ndarray) -> np.ndarray:
    r"""evaluates the image Laplacian for each of the D maps, :math:`\Delta \Theta_{\cdot d}`

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        D vectors of N pixels
    list_edges : np.ndarray
        set of edges in the graph induced by the spatial regularization

    Returns
    -------
    np.ndarray of shape (N, D)
        image Laplacian for each of the D maps
    """
    laplacian_ = np.zeros_like(Theta)
    N = laplacian_.shape[0]

    for i in range(N):
        mask_i_m = list_edges[:, 1] == i
        mask_i_p = list_edges[:, 0] == i

        laplacian_[i] += np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :]),
            axis=0,
        )
        laplacian_[i] -= np.sum(
            (Theta[list_edges[mask_i_m, 1], :] - Theta[list_edges[mask_i_m, 0], :]),
            axis=0,
        )
    return laplacian_  # (N, D)


@numba.njit()
def compute_laplacian_local(
    Theta: np.ndarray, n: int, list_edges: np.ndarray, list_pixel_candidates: np.ndarray
) -> np.ndarray:
    """computes the local laplacian only for the one modified pixel

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        current iterate

    n : int
        the index of the pixel to consider (:math:`0 \leq n \leq N - 1`)

    list_pixel_candidates : np.ndarray of shape (N_candidates, D)
        the list of all candidates for pixel n

    Returns
    -------
    laplacian_ : np.ndarray of shape (N_candidates, D)
        the laplacian of the candidates
    """
    # D = Theta.shape[1]
    N_candidates = list_pixel_candidates.shape[0]
    laplacian_ = np.zeros_like(list_pixel_candidates)

    mask_i_m = list_edges[:, 1] == n
    mask_i_p = list_edges[:, 0] == n

    for i in range(N_candidates):
        Theta[n] = list_pixel_candidates[i] * 1

        laplacian_[i] += np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :]),
            axis=0,
        )  # (D,)
        laplacian_[i] -= np.sum(
            (Theta[list_edges[mask_i_m, 1], :] - Theta[list_edges[mask_i_m, 0], :]),
            axis=0,
        )  # (D,)
    return laplacian_  # (N_candidates, D)


@numba.njit()
def compute_gradient_from_laplacian(
    laplacian_: np.ndarray, list_edges: np.ndarray
) -> np.ndarray:
    """evaluates the gradient from the Laplacian matrix

    Parameters
    ----------
    laplacian_ : _type_
        _description_
    list_edges : _type_
        _description_

    Returns
    -------
    np.ndarray
        _description_
    """
    g = np.zeros_like(laplacian_)

    for edge in list_edges:
        val = 2 * (laplacian_[edge[1]] - laplacian_[edge[0]])  # (D,)
        g[edge[0]] += val
        g[edge[1]] -= val

    return g  # (N, D)


@numba.njit()
def _neglog_pdf_one_pix(
    Theta: np.ndarray,
    idx_pix: np.ndarray,
    list_pixel_candidates: np.ndarray,
    list_edges: np.ndarray,
) -> np.ndarray:
    """computes the neg log-prior when only one pixel is modified

    Parameters
    ----------
    Theta : np.ndarray of shape (N, D)
        current iterate

    idx_pix : np.ndarray
        array of the indices of the pixels to consider (:math:`0 \leq n \leq N - 1`)

    list_pixel_candidates : np.ndarray of shape (N_candidates, D)
        the list of all candidates for pixel n

    Returns
    -------
    np.ndarray of shape (N_candidates,)
        the leg log-prior of the candidates
    """
    n_pix, k_mtm, D = list_pixel_candidates.shape
    neglog_p = np.zeros((n_pix, k_mtm, D))
    i = 0
    n_previous = -500_000

    for n in idx_pix:
        if n != n_previous:
            list_edges_pix = list_edges[
                (list_edges[:, 0] == n) | (list_edges[:, 1] == n)
            ]

        if list_edges_pix.size > 0:
            laplacian_ = compute_laplacian_local(
                Theta, n, list_edges_pix, list_pixel_candidates[i]
            )  # (k_mtm, D)
            neglog_p[i, :, :] += laplacian_**2  # (n_pix, k_mtm, D)

        i += 1
        n_previous = n * 1

    return neglog_p


class L22LaplacianSpatialPrior(SpatialPrior):
    r"""L22 smooth spatial prior, valid for both 1D and 2D tensors. Its pdf is defined as

    .. math::

        \forall d \in [1, D], \quad \pi(\Theta_{\cdot d}) \propto \exp \left[- \tau_d \Vert \Delta \Theta_{\cdot d} \Vert_F^2 \right]

    where  :math:`\Vert \cdot \Vert_F` denotes the Fröbenius norm and :math:`\Delta \Theta_{\cdot d}` is the Laplacian of vector :math:ù\Theta_{\cdot d}`.
    """

    def neglog_pdf(
        self,
        Theta: np.ndarray,
        with_weights: bool = True,
        pixelwise: bool = False,
    ) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        if pixelwise:
            neglog_p = np.zeros((self.N, self.D))
        else:
            neglog_p = np.zeros((self.D,))

        if self.list_edges.size > 0:
            laplacian_ = compute_laplacian(Theta, self.list_edges)
            if pixelwise:
                neglog_p += laplacian_**2  # (N,D)
            else:
                neglog_p += np.sum(laplacian_**2, axis=0)  # (D,)

        if with_weights:
            if pixelwise:
                neglog_p = self.weights[None, :] * neglog_p
            else:
                neglog_p *= self.weights

        # neglog_p /= self.N * self.D
        return neglog_p  # (D,) if not pixelwise or (N, D) if pixelwise

    def neglog_pdf_one_pix(
        self,
        Theta: np.ndarray,
        idx_pix: np.ndarray,
        list_pixel_candidates: np.ndarray,
        other_weights: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        computes the neg log-prior when only one pixel is modified

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            current iterate

        idx_pix : np.ndarray
            array of the indices of the pixels to consider (:math:`0 \leq n \leq N - 1`)

        list_pixel_candidates : np.ndarray of shape (N_candidates, D)
            the list of all candidates for pixel n

        Returns
        -------
        np.ndarray of shape (N_candidates,)
            the leg log-prior of the candidates
        """
        neglog_p = _neglog_pdf_one_pix(
            Theta, idx_pix, list_pixel_candidates, self.list_edges
        )  # # (n_pix, k_mtm, D)
        # neglog_p /= self.N * self.D

        if other_weights is None:
            return np.sum(
                neglog_p * self.weights[None, None, :], axis=2
            )  # (n_pix, k_mtm)
        else:
            return np.sum(
                neglog_p * other_weights[None, None, :], axis=2
            )  # (n_pix, k_mtm)

    def gradient_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        laplacian_ = compute_laplacian(Theta, self.list_edges)

        g = compute_gradient_from_laplacian(laplacian_, self.list_edges)  # (N, D)
        # g /= self.N * self.D
        return self.weights[None, :] * g  # (N, D)

    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        hess_diag = np.zeros_like(Theta, dtype=np.float64)

        if self.list_edges.size > 0:
            idx, counts = np.unique(self.list_edges.flatten(), return_counts=True)
            # print(counts.dtype, hess_diag.dtype)
            hess_diag[idx, :] += (
                2 * (counts * (counts + 1))[:, None] * np.ones((idx.size, self.D))
            )

        # hess_diag /= self.N * self.D
        return self.weights[None, :] * hess_diag  # (N, D)



"""Marginal maximum likelihood update for scalar regularization parameters.
"""
from abc import ABC, abstractmethod

import numpy as np
from numba import jit


@jit(nopython=True, cache=True)
def pgd_rate(theta, stepsize, vmin, vmax, dim, gX, homogeneity):
    r"""Update parameter :math:`\theta` from iteration :math:`k` to
    :math:`k+1` using projected gradient ascent, assuming
    :math:`p(x \mid \theta)` is of the form

    .. math::
        p(x \mid \theta) \propto \exp(- \theta g(x)),

    with :math:`g \in \Gamma_0(\mathbb{R^N})`.

    Parameters
    ----------
    theta : float
        Current value of the rate parameter :math:`\theta^{(k)}` to be
        optimized.
    stepsize : float
        Stepsize.
    vmin : float
        Lower limit of the admissible interval.
    vmax : float
        Upper limit of the admissible interval.
    dim : float
        Dimension of the space over which the distribution
            :math:`p(x \mid \theta)` is defined.
    gX : float
        Value of the potential function :math:`g` evaluated using the
        current iterate.
    homogeneity : float
        Positive homogeneity factor of the function
        :math:`- \log p(x \mid \theta)` with respect to :math:`x`. By
        default 1.

    Returns
    -------
    float
        Updated value :math:`\theta^{(k+1)}`.
    """
    theta = np.maximum(
        np.minimum(
            theta + stepsize * (dim / (homogeneity * theta) - gX),
            vmax,
        ),
        vmin,
    )
    return theta


@jit(nopython=True, cache=True)
def pgd_scale(theta, stepsize, vmin, vmax, dim, gX, homogeneity):
    r"""Update parameter :math:`\theta` from iteration :math:`k` to
    :math:`k+1` using projected gradient ascent, assuming
    :math:`p(x \mid \theta)` is of the form

    .. math::
        p(x \mid \theta) \propto \exp(- g(x) / \theta),

    with :math:`g \in \Gamma_0(\mathbb{R^N})`.

    Parameters
    ----------
    theta : float
        Current value of the scale parameter :math:`\theta^{(k)}` to be
        optimized.
    stepsize : float
        Stepsize.
    vmin : float
        Lower limit of the admissible interval.
    vmax : float
        Upper limit of the admissible interval.
    dim : float
        Dimension of the space over which the distribution
            :math:`p(x \mid \theta)` is defined.
    gX : float
        Value of the potential function :math:`g` evaluated using the
        current iterate.
    homogeneity : float
        Positive homogeneity factor of the function
        :math:`- \log p(x \mid \theta)` with respect to :math:`x`. By
        default 1.

    Returns
    -------
    float
        Updated value :math:`\theta^{(k+1)}`.
    """
    theta = np.maximum(
        np.minimum(
            theta + (stepsize / theta) * (-dim / homogeneity + gX / theta),
            vmax,
        ),
        vmin,
    )
    return theta


@jit(nopython=True, cache=True)
def pgd_rate_log(theta, stepsize, vmin, vmax, dim, gX, homogeneity):
    r"""Update parameter :math:`\log \theta^{(k)}` from iteration :math:`k` to
    :math:`k+1` using projected gradient ascent (in log scale), assuming
    :math:`p(x \mid \theta)` is of the form

    .. math::
        p(x \mid \theta) \propto \exp(- \theta g(x)),

    with :math:`g \in \Gamma_0(\mathbb{R^N})`.

    Parameters
    ----------
    theta : float
        Current value of the rate parameter :math:`\theta^{(k)}` to be
        optimized.
    stepsize : float
        Stepsize.
    vmin : float
        Lower limit of the admissible interval for :math:`\log \theta`.
    vmax : float
        Upper limit of the admissible interval for :math:`\log \theta`.
    dim : float
        Dimension of the space over which the distribution
        :math:`p(x \mid \theta)` is defined.
    gX : float
        Value of the potential function :math:`g` evaluated using the
        current iterate.
    homogeneity : float
        Positive homogeneity factor of the function
        :math:`- \log p(x \mid \theta)` with respect to :math:`x`. By
        default 1.

    Returns
    -------
    float
        Updated value :math:`\theta^{(k+1)}` (in linear scale).
    """
    log_theta = np.log(theta)
    theta = np.exp(
        np.maximum(
            np.minimum(
                log_theta + stepsize * (dim / homogeneity - gX * theta),
                vmax,
            ),
            vmin,
        )
    )
    return theta


@jit(nopython=True, cache=True)
def pgd_scale_log(theta, stepsize, vmin, vmax, dim, gX, homogeneity):
    r"""Update parameter :math:`\theta` from iteration :math:`k` to
    :math:`k+1` using projected gradient ascent (in log scale), assuming
    :math:`p(x \mid \theta)` is of the form

    .. math::
        p(x \mid \theta) \propto \exp(- g(x) / \theta),

    with :math:`g \in \Gamma_0(\mathbb{R^N})`.

    Parameters
    ----------
    theta : float
        Current value of the scale parameter :math:`\theta^{(k)}` to be
        optimized.
    stepsize : float
        Stepsize.
    vmin : float
        Lower limit of the admissible interval.
    vmax : float
        Upper limit of the admissible interval.
    dim : float
        Dimension of the space over which the distribution
            :math:`p(x \mid \theta)` is defined.
    gX : float
        Value of the potential function :math:`g` evaluated using the
        current iterate.
    homogeneity : float
        Positive homogeneity factor of the function
        :math:`- \log p(x \mid \theta)` with respect to :math:`x`. By
        default 1.

    Returns
    -------
    float
        Updated value :math:`\theta^{(k+1)}` (in linear scale).
    """
    log_theta = np.log(theta)
    theta = np.exp(
        np.maximum(
            np.minimum(
                log_theta + stepsize * (-dim / homogeneity + gX / theta),
                vmax,
            ),
            vmin,
        )
    )
    return theta


class EBayesMMLE(ABC):
    r"""Abstract class implementing an empirical Bayes maximum marginal
    likelihood approach to estimate a scalar regularization parameter.

    Empirical Bayes maximum marginal likelihood estimation of a scalar
    regularization parameter :math:`\theta` from within an MCMC algorithm using
    a projected gradient ascent. The implementation follows the description
    provided in :cite:t:`vidalMaximumLikelihoodEstimation2020`.
    """
    # :cite:p:`Vidal2020`

    __slots__ = (
        "scale",
        "N0",
        "N1",
        "dim",
        "vmin",
        "vmax",
        "homogeneity",
        "exponent",
        "_c",
        "mean_theta",
        "mean_theta_old",
        "sum_weights",
    )

    def __init__(
        self,
        scale,
        N0,
        N1,
        dim,
        vmin,
        vmax,
        homogeneity=1.0,
        exponent=0.8,
    ):
        r"""BayesMMLE constructor.

        Parameters
        ----------
        scale : float
            Scale parameter involved in the definition of the projected gradient
            stepsize.
        N0 : int
            Number of iterations defining the initial update phase.
        N1 : int
            Number of iterations for the stabilization phase.
        dim : int
            Dimension of the space over which the distribution
            :math:`p(x \mid \theta)` is defined.
        vmin : float
            Lower limit of the admissible interval.
        vmax : float
            Upper limit of the admissible interval.
        homogeneity : float, optional
            Positive homogeneity factor of the function
            :math:`- \log p(x \mid \theta)` with respect to :math:`x`. By
            default 1.
        exponent : float, optional
            Exponent involved in the evolution of the stepsize of the
            projected gradient. By default 0.8
        """
        self.scale = scale
        self.N0 = N0
        self.N1 = N1
        self.dim = dim
        self.vmin = vmin
        self.vmax = vmax
        self.homogeneity = homogeneity
        self.exponent = exponent
        self._c = scale / dim
        self.mean_theta = 0.0
        self.mean_theta_old = 0.0
        self.sum_weights = 0.0

    def _compute_stepsize(self, iteration):
        r"""Compute current value for the stepsize of the projected gradient.

        Parameters
        ----------
        iteration : int
            Current iteration of the sampler.

        Returns
        -------
        float
            Current value for the stepsize.
        """
        stepsize = self._c / ((iteration - self.N0 + 1) ** self.exponent)
        return stepsize

    def _update_mean_theta(self, theta, iteration, stepsize):
        """Compute average value for the parameter.

        Parameters
        ----------
        theta : float
            Current state of the parameter.
        iteration : int
            Current iteration index of the sampler.
        stepsize : float
            Stepsize for the update of the parameter investigated.
        """
        self.mean_theta_old = self.mean_theta
        if iteration > self.N1:
            step = stepsize  # phase 3: decreasing stepsize (weights)
        else:
            step = 1  # phase 2: uniform weights
        self.mean_theta = self.sum_weights * self.mean_theta + step * theta
        self.sum_weights += step
        self.mean_theta /= self.sum_weights

    @staticmethod
    @abstractmethod
    def _update_step(theta, stepsize, vmin, vmax, dim, gX, homogeneity):
        return NotImplemented

    def update(self, theta, iteration, gX):
        r"""Update parameter :math:`\theta` from iteration :math:`k` to
        :math:`k+1`.

        Parameters
        ----------
        theta : float
            Current value of the parameter to be optimized.
        iteration : int
            Current iteration of the sampler.
        gX : float
            Value of the potential function :math:`g` evaluated using the
            current iterate.

        Returns
        -------
        float
            Updated value \theta^{(k+1)}.
        """
        if iteration + 1 > self.N0:  # no update if not in phase > 1
            stepsize = self._compute_stepsize(iteration)
            theta = self._update_step(
                theta,
                stepsize,
                self.vmin,
                self.vmax,
                self.dim,
                gX,
                self.homogeneity,
            )
            self._update_mean_theta(theta, iteration, stepsize)
        return theta


class EBayesMMLERate(EBayesMMLE):
    r"""Empirical Bayes maximum marginal likelihood approach to estimate a
    parameter :math:`\theta` in linear scale when

        .. math::
            p(x, y \mid \theta) \propto p(y \mid x) p(x \mid \theta)
            p(x \mid \theta) = Z^{-1}(\theta) \exp(-\theta g(x))

    with :math:`g` a proper, closed, convex and :math:`\alpha`-positively
    homogeneous function.
    """
    _update_step = staticmethod(pgd_rate)


class EBayesMMLELogRate(EBayesMMLE):
    r"""Empirical Bayes maximum marginal likelihood approach to estimate a
    parameter :math:`\theta` in log scale when

        .. math::
            p(x, y \mid \theta) \propto p(y \mid x) p(x \mid \theta)
            p(x \mid \theta) = Z^{-1}(\theta) \exp(-\theta g(x))

    with :math:`g` a proper, closed, convex and :math:`\alpha`-positively
    homogeneous function.
    """

    def __init__(
        self,
        scale,
        N0,
        N1,
        dim,
        vmin=1e-8,
        vmax=1e8,
        homogeneity=1.0,
        exponent=0.8,
    ):
        super(EBayesMMLELogRate, self).__init__(
            scale,
            N0,
            N1,
            dim,
            np.log(vmin),
            np.log(vmax),
            homogeneity=homogeneity,
            exponent=exponent,
        )

    _update_step = staticmethod(pgd_rate_log)


class EBayesMMLEScale(EBayesMMLE):
    r"""Empirical Bayes maximum marginal likelihood approach to estimate a
    parameter :math:`\theta` in linear scale when

        .. math::
            p(x, y \mid \theta) \propto p(y \mid x) p(x \mid \theta)
            p(x \mid \theta) = Z^{-1}(\theta) \exp(-\theta^{-1} g(x))

    with :math:`g` a proper, closed, convex and :math:`\alpha`-positively
    homogeneous function.
    """
    _update_step = staticmethod(pgd_scale)


class EBayesMMLELogScale(EBayesMMLE):
    r"""Empirical Bayes maximum marginal likelihood approach to estimate a
    parameter :math:`\theta` in log scale when

        .. math::
            p(x, y \mid \theta) \propto p(y \mid x) p(x \mid \theta)
            p(x \mid \theta) = Z^{-1}(\theta) \exp(-\theta^{-1} g(x))

    with :math:`g` a proper, closed, convex and :math:`\alpha`-positively
    homogeneous function.
    """

    def __init__(
        self,
        scale,
        N0,
        N1,
        dim,
        vmin,
        vmax,
        homogeneity=1.0,
        exponent=0.8,
    ):
        super(EBayesMMLELogScale, self).__init__(
            scale,
            N0,
            N1,
            dim,
            np.log(vmin),
            np.log(vmax),
            homogeneity=homogeneity,
            exponent=exponent,
        )

    _update_step = staticmethod(pgd_scale_log)
